from __future__ import annotations

import functools
import pathlib
from abc import ABC
from dataclasses import dataclass
from typing import Literal, Mapping, Sequence, Union, overload

import ase
import ase.db
import ase.io
import locache
import torch.utils.data
from load_atoms import load_dataset

from graph_pes.atomic_graph import (
    ALL_PROPERTY_KEYS,
    AtomicGraph,
    PropertyKey,
)
from graph_pes.data.ase_db import ASEDatabase
from graph_pes.utils.logger import logger
from graph_pes.utils.misc import MultiSequence, slice_to_range, uniform_repr
from graph_pes.utils.sampling import SequenceSampler


class GraphDataset(torch.utils.data.Dataset, ABC):
    """
    A dataset of :class:`~graph_pes.AtomicGraph` instances.

    Parameters
    ----------
    graphs
        The collection of :class:`~graph_pes.AtomicGraph` instances.
    """

    def __init__(self, graphs: Sequence[AtomicGraph]):
        self.graphs = graphs
        # raise errors on instantiation if accessing a datapoint would fail
        _ = self[0]

    def __getitem__(self, index: int) -> AtomicGraph:
        return self.graphs[index]

    def __len__(self) -> int:
        return len(self.graphs)

    def prepare_data(self) -> None:
        """
        Make general preparations for loading the data for the dataset.

        Called on rank-0 only: don't set any state here.
        May be called multiple times.
        """

    def setup(self) -> None:
        """
        Set-up the data for this specific instance of the dataset.

        Called on every process in the distributed setup. May be called
        multiple times.
        """

    @property
    def properties(self) -> list[PropertyKey]:
        """The properties that are available to train on with this dataset"""

        # assume all data points have the same labels
        example_graph = self[0]
        return [
            key for key in ALL_PROPERTY_KEYS if key in example_graph.properties
        ]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({len(self):,}, "
            f"properties={self.properties})"
        )


class ConcatDataset(GraphDataset):
    """
    A dataset that concatenates multiple :class:`GraphDataset` instances.
    Useful for e.g. training on datasets from multiple files simultaneously:

    .. code-block:: yaml

        data:
            train:
                +ConcatDataset:
                    dimers:
                        +file_dataset:
                            path: dimers.xyz
                            cutoff: 5.0
                    crystals:
                        +file_dataset:
                            path: crystals.xyz
                            cutoff: 5.0

            valid:
                ...


    Parameters
    ----------
    datasets
        The collection of :class:`GraphDataset` instances to concatenate.
        The keys are arbitrary names for the datasets, and the values are
        the :class:`GraphDataset` instances.
    """

    def __init__(self, **datasets: GraphDataset):
        self.datasets = datasets
        self._lengths = {k: len(v) for k, v in datasets.items()}

    def __len__(self):
        return sum(self._lengths.values())

    def __getitem__(self, index: int) -> AtomicGraph:
        for k, v in self._lengths.items():
            if index < v:
                return self.datasets[k][index]
            index -= v
        raise IndexError(f"Index {index} is out of bounds for the dataset")

    def prepare_data(self) -> None:
        for dataset in self.datasets.values():
            dataset.prepare_data()

    def setup(self) -> None:
        for dataset in self.datasets.values():
            dataset.setup()

    @property
    def properties(self) -> list[PropertyKey]:
        return list(
            set(
                p
                for dataset in self.datasets.values()
                for p in dataset.properties
            )
        )

    @property
    def graphs(self) -> MultiSequence[AtomicGraph]:
        return MultiSequence(
            [dataset.graphs for dataset in self.datasets.values()],
        )


class ASEToGraphsConverter(Sequence[AtomicGraph]):
    def __init__(
        self,
        structures: Sequence[ase.Atoms],
        cutoff: float,
        property_mapping: Mapping[str, PropertyKey] | None = None,
        others_to_include: list[str] | None = None,
    ):
        self.structures = structures
        self.cutoff = cutoff
        self.property_mapping = property_mapping
        self.others_to_include = others_to_include

    @overload
    def __getitem__(self, index: int) -> AtomicGraph: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[AtomicGraph]: ...
    def __getitem__(
        self, index: int | slice
    ) -> AtomicGraph | Sequence[AtomicGraph]:
        if isinstance(index, slice):
            indices = slice_to_range(index, len(self))
            return [self[i] for i in indices]

        return AtomicGraph.from_ase(
            self.structures[index],
            cutoff=self.cutoff,
            property_mapping=self.property_mapping,
            others_to_include=self.others_to_include,
        )

    def __len__(self) -> int:
        return len(self.structures)


# use the locache library to cache the graphs that result from this
# transform to disk: this means that multiple training runs on the
# same dataset will be able to reuse the same graphs, massively speeding
# up the start to training for the (n>1)th run
# to ensure that the graphs we get back are of the correct dtype,
# we need to pass the current torch dtype to this caching function
@locache.persist
def get_all_graphs_and_cache_to_disk(
    converter: ASEToGraphsConverter,
    torch_dtype: torch.dtype,
) -> list[AtomicGraph]:
    logger.info(
        f"Caching neighbour lists for {len(converter)} structures "
        f"with cutoff {converter.cutoff}, property mapping "
        f"{converter.property_mapping} and torch dtype {torch_dtype}"
    )
    return [converter[i] for i in range(len(converter))]


class ASEToGraphDataset(GraphDataset):
    """
    A dataset that wraps a :class:`Sequence` of :class:`ase.Atoms`, and converts
    them to :class:`~graph_pes.AtomicGraph` instances.

    Parameters
    ----------
    structures
        The collection of :class:`ase.Atoms` objects to convert to
        :class:`~graph_pes.AtomicGraph` instances.
    cutoff
        The cutoff to use when creating neighbour indexes for the graphs.
    pre_transform
        Whether to precompute the the :class:`~graph_pes.AtomicGraph`
        objects, or only do so on-the-fly when the dataset is accessed.
        This pre-computations stores the graphs in memory, and so will be
        prohibitively expensive for large datasets.
    property_mapping
        A mapping from properties defined on the :class:`ase.Atoms` objects to
        their appropriate names in ``graph-pes``, see
        :meth:`~graph_pes.AtomicGraph.from_ase`.
    others_to_include
        A list of properties to include in the ``graph.other`` field
        that are present as per-atom or per-structure properties on the
        :class:`ase.Atoms` objects.
    """

    def __init__(
        self,
        structures: Sequence[ase.Atoms],
        cutoff: float,
        pre_transform: bool = False,
        property_mapping: Mapping[str, PropertyKey] | None = None,
        others_to_include: list[str] | None = None,
    ):
        super().__init__(
            ASEToGraphsConverter(
                structures, cutoff, property_mapping, others_to_include
            ),
        )
        self.pre_transform = pre_transform

    def prepare_data(self):
        if self.pre_transform:
            # cache the graphs to disk - this is done on rank-0 only
            # and means that expensive data pre-transforms don't need to be
            # recomputed on each rank in the distributed setup
            get_all_graphs_and_cache_to_disk(
                self.graphs, torch.get_default_dtype()
            )

    def setup(self) -> None:
        if self.pre_transform and isinstance(self.graphs, ASEToGraphsConverter):
            # load the graphs from disk
            actual_graphs = get_all_graphs_and_cache_to_disk(
                self.graphs, torch.get_default_dtype()
            )
            self.graphs = actual_graphs


@dataclass
class DatasetCollection:
    """
    A convenience container for training, validation, and optional test sets.
    """

    train: GraphDataset
    """The training dataset."""
    valid: GraphDataset
    """The validation dataset."""
    test: Union[GraphDataset, dict[str, GraphDataset], None] = None  # noqa: UP007
    """An optional test dataset, or collection of named test datasets."""

    def __repr__(self) -> str:
        kwargs = {"train": self.train, "valid": self.valid}
        if self.test is not None:
            kwargs["test"] = self.test  # type: ignore
        return uniform_repr(
            self.__class__.__name__,
            **kwargs,  # type: ignore
        )


def load_atoms_dataset(
    id: str | pathlib.Path,
    cutoff: float,
    n_train: int,
    n_valid: int,
    n_test: int | None = None,
    split: Literal["random", "sequential"] = "random",
    seed: int = 42,
    pre_transform: bool = True,
    property_map: dict[str, PropertyKey] | None = None,
    others_to_include: list[str] | None = None,
) -> DatasetCollection:
    """
    Load an dataset of :class:`ase.Atoms` objects using
    `load-atoms <https://jla-gardner.github.io/load-atoms/>`__,
    convert them to :class:`~graph_pes.AtomicGraph` instances, and split into
    train and valid sets.

    Examples
    --------
    Load a subset of the QM9 dataset. Ensure that the ``U0`` property is
    mapped to ``energy``:

    >>> load_atoms_dataset(
    ...     "QM9",
    ...     cutoff=5.0,
    ...     n_train=1_000,
    ...     n_valid=100,
    ...     n_test=100,
    ...     property_map={"U0": "energy"},
    ... )

    Use this to specify a complete collection of datasets (train, val and test)
    in a YAML configuration file:

    .. code-block:: yaml

        data:
            +load_atoms_dataset:
                id: QM9
                cutoff: 5.0
                n_train: 1_000
                n_valid: 100
                n_test: 100
                property_map:
                    U0: energy

    Parameters
    ----------
    id:
        The dataset identifier. Can be a ``load-atoms`` id, or a path to an
        ``ase``-readable data file.
    cutoff:
        The cutoff radius for the neighbor list.
    n_train:
        The number of training structures.
    n_valid:
        The number of validation structures.
    n_test:
        The number of test structures. If ``None``, no test set is created.
    split:
        The split method. ``"random"`` shuffles the structures before
        choosing a non-overlapping split, while ``"sequential"`` takes the
        first ``n_train`` structures for training and the next ``n_valid``
        structures for validation.
    seed:
        The random seed.
    pre_transform:
        Whether to pre-calculate the neighbour lists for each structure.
    root:
        The root directory
    property_map:
        A mapping from properties as named on the atoms objects to
        ``graph-pes`` property keys, e.g. ``{"U0": "energy"}``.
    others_to_include:
        A list of properties to include in the ``graph.other`` field
        that are present as per-atom or per-structure properties on the
        :class:`ase.Atoms` objects.

    Returns
    -------
    DatasetCollection
        A collection of training, validation, and optional test datasets.
    """
    _dataset_factory = functools.partial(
        ASEToGraphDataset,
        cutoff=cutoff,
        pre_transform=pre_transform,
        property_mapping=property_map,
        others_to_include=others_to_include,
    )

    structures = SequenceSampler(load_dataset(id))

    if split == "random":
        structures = structures.shuffled(seed)

    train = structures[:n_train]
    val = structures[n_train : n_train + n_valid]

    return DatasetCollection(
        train=_dataset_factory(train),
        valid=_dataset_factory(val),
        test=_dataset_factory(
            structures[n_train + n_valid : n_train + n_valid + n_test],
        )
        if n_test is not None
        else None,
    )


def file_dataset(
    path: str | pathlib.Path,
    cutoff: float,
    n: int | None = None,
    shuffle: bool = True,
    seed: int = 42,
    pre_transform: bool = True,
    property_map: dict[str, PropertyKey] | None = None,
    others_to_include: list[str] | None = None,
) -> ASEToGraphDataset:
    """
    Load an ASE dataset from a file that is either:

    * any plain-text file that can be read by :func:`ase.io.read`, e.g. an
      ``.xyz`` file
    * a ``.db`` file containing a SQLite database of :class:`ase.Atoms` objects
      that is readable as an `ASE database <https://wiki.fysik.dtu.dk/ase/ase/db/db.html>`__.
      Under the hood, this uses the :class:`~graph_pes.data.ase_db.ASEDatabase`
      class - see there for more details.

    Examples
    --------
    Load a dataset from a file, ensuring that the ``energy`` property is
    mapped to ``U0``:

    >>> file_dataset(
    ...     "training_data.xyz",
    ...     cutoff=5.0,
    ...     property_map={"U0": "energy"},
    ... )

    Use this to specify e.g. a training dataset in a YAML configuration file:

    .. code-block:: yaml

        data:
            train:
                +file_dataset:
                    path: training_data.xyz
                    cutoff: 5.0
                    property_map:
                    U0: energy

    Parameters
    ----------
    path:
        The path to the file.
    cutoff:
        The cutoff radius for the neighbour list.
    n:
        The number of structures to load. If ``None``,
        all structures are loaded.
    shuffle:
        Whether to shuffle the structures.
    seed:
        The random seed used for shuffling.
    pre_transform:
        Whether to pre-calculate the neighbour lists for each structure.
    property_map:
        A mapping from properties as named on the atoms objects to
        ``graph-pes`` property keys, e.g. ``{"U0": "energy"}``.
    others_to_include:
        A list of properties to include in the ``graph.other`` field
        that are present as per-atom or per-structure properties on the
        :class:`ase.Atoms` objects.

    Returns
    -------
    ASEToGraphDataset
        The ASE dataset.
    """

    if isinstance(path, str):
        path = pathlib.Path(path)

    if path.suffix == ".db":
        structures = ASEDatabase(path)
    else:
        structures = ase.io.read(path, index=":")
        assert isinstance(structures, list)

    structure_collection = SequenceSampler(structures)
    if shuffle:
        structure_collection = structure_collection.shuffled(seed)

    if n is not None:
        structure_collection = structure_collection[:n]

    return ASEToGraphDataset(
        structure_collection,
        cutoff,
        pre_transform,
        property_map,
        others_to_include,
    )



# ruff: noqa: UP006, UP007
# ^^ NB: dacite parsing requires the old type hint syntax in
#        order to be compatible with all versions of Python that
#        we are targeting (3.9+)
from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Literal, Union

import yaml
from pytorch_lightning import Callback

from graph_pes.config.shared import TorchConfig
from graph_pes.data.datasets import DatasetCollection
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.training.callbacks import VerboseSWACallback
from graph_pes.training.loss import Loss, TotalLoss
from graph_pes.training.opt import LRScheduler, Optimizer


@dataclass
class EarlyStoppingConfig:
    patience: int
    """
    The number of validation checks with no improvement before stopping.
    """

    min_delta: float = 0.0
    """
    The minimum change in the monitored quantity to qualify as an improvement.
    """

    monitor: str = "valid/loss/total"
    """The quantity to monitor."""


@dataclass
class FittingOptions:
    """Options for the fitting process."""

    pre_fit_model: bool
    max_n_pre_fit: Union[int, None]
    early_stopping: Union[EarlyStoppingConfig, None]
    loader_kwargs: Dict[str, Any]
    early_stopping_patience: Union[int, None]
    """
    DEPRECATED: use the `early_stopping` config option instead.
    """


@dataclass
class SWAConfig:
    """
    Configuration for Stochastic Weight Averaging.

    Internally, this is handled by `this PyTorch Lightning callback
    <https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.callbacks.StochasticWeightAveraging.html>`__.
    """

    lr: float
    """
    The learning rate to use during the SWA phase. If not specified,
    the learning rate from the end of the training phase will be used.
    """

    start: Union[int, float] = 0.8
    """
    The epoch at which to start SWA. If a float, it will be interpreted
    as a fraction of the total number of epochs.
    """

    anneal_epochs: int = 10
    """
    The number of epochs over which to linearly anneal the learning rate
    to zero.
    """

    strategy: Literal["linear", "cos"] = "linear"
    """The strategy to use for annealing the learning rate."""

    def instantiate_lightning_callback(self):
        return VerboseSWACallback(
            swa_lrs=self.lr,
            swa_epoch_start=self.start,
            annealing_epochs=self.anneal_epochs,
            annealing_strategy=self.strategy,
        )


@dataclass
class FittingConfig(FittingOptions):
    """Configuration for the fitting process."""

    trainer_kwargs: Dict[str, Any]
    optimizer: Optimizer = Optimizer(name="AdamW", lr=1e-3, amsgrad=False)
    scheduler: Union[LRScheduler, None] = None
    swa: Union[SWAConfig, None] = None
    callbacks: List[Callback] = field(default_factory=list)


@dataclass
class GeneralConfig:
    """General configuration for a training run."""

    seed: int
    root_dir: str
    run_id: Union[str, None]
    torch: TorchConfig
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO"
    progress: Literal["rich", "logged"] = "rich"


# TODO:
# - move get_model to utils, call it parse_model
# - turn loss into dict with human readable names + move parse_loss to utils


@dataclass
class TrainingConfig:
    """
    A schema for a configuration file to train a
    :class:`~graph_pes.GraphPESModel`.
    """

    model: Union[GraphPESModel, Dict[str, GraphPESModel]]
    data: DatasetCollection
    loss: Union[Loss, TotalLoss, Dict[str, Loss], List[Loss]]
    fitting: FittingConfig
    general: GeneralConfig
    wandb: Union[Dict[str, Any], None]

    ### Methods ###

    def get_data(self) -> DatasetCollection:
        if isinstance(self.data, DatasetCollection):
            return self.data
        elif isinstance(self.data, dict):
            return DatasetCollection(**self.data)

        raise ValueError(
            "Expected to be able to parse a DatasetCollection instance or a "
            "dictionary mapping 'train' and 'valid' keys to GraphDataset "
            "instances from the data config, but got something else: "
            f"{self.data}"
        )

    @classmethod
    def defaults(cls) -> dict:
        with open(Path(__file__).parent / "training-defaults.yaml") as f:
            return yaml.safe_load(f)



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Final, Sequence, final

import torch
from ase.data import chemical_symbols
from torch import nn

from graph_pes.atomic_graph import (
    AtomicGraph,
    PropertyKey,
    has_cell,
    is_batch,
    replace,
    sum_per_structure,
    to_batch,
    trim_edges,
)
from graph_pes.data.datasets import GraphDataset
from graph_pes.utils.logger import logger

from .utils.misc import differentiate, differentiate_all
from .utils.nn import PerElementParameter

if TYPE_CHECKING:
    from graph_pes.utils.calculator import GraphPESCalculator


class GraphPESModel(nn.Module, ABC):
    r"""
    All models implemented in ``graph-pes`` are subclasses of
    :class:`~graph_pes.GraphPESModel`.

    These models make predictions (via the
    :meth:`~graph_pes.GraphPESModel.predict` method) of the
    following properties:

    .. list-table::
            :header-rows: 1

            * - Key
              - Single graph
              - Batch of graphs
              - Units
            * - :code:`"local_energies"`
              - :code:`(N,)`
              - :code:`(N,)`
              - :code:`[energy]`
            * - :code:`"energy"`
              - :code:`()`
              - :code:`(M,)`
              - :code:`[energy]`
            * - :code:`"forces"`
              - :code:`(N, 3)`
              - :code:`(N, 3)`
              - :code:`[energy / length]`
            * - :code:`"stress"`
              - :code:`(3, 3)`
              - :code:`(M, 3, 3)`
              - :code:`[energy / length^3]`
            * - :code:`"virial"`
              - :code:`(3, 3)`
              - :code:`(M, 3, 3)`
              - :code:`[energy]`

    assuming an input of an :class:`~graph_pes.AtomicGraph` representing a
    single structure composed of ``N`` atoms, or an
    :class:`~graph_pes.AtomicGraph` composed of ``M`` structures and containing
    a total of ``N`` atoms. (see :func:`~graph_pes.atomic_graph.is_batch` for
    more information about batching).

    Note that ``graph-pes`` makes no assumptions as to the actual units of
    the ``energy`` and ``length`` quantities - these will depend on the
    labels the model has been trained on (e.g. could be ``eV`` and ``Å``,
    ``kcal/mol`` and ``nm`` or even ``J`` and ``m``).

    Implementations must override the
    :meth:`~graph_pes.GraphPESModel.forward` method to generate a
    dictionary of predictions for the given graph. As a minimum, this must
    include a per-atom energy contribution (``"local_energies"``).

    For any other properties not returned by the forward pass,
    the :meth:`~graph_pes.GraphPESModel.predict` method will automatically
    infer these properties from the local energies as required:

    * ``"energy"``: as the sum of the local energies per structure.
    * ``"forces"``: as the negative gradient of the energy with respect to the
      atomic positions.
    * ``"stress"``: as the negative gradient of the energy with respect to a
      symmetric expansion of the unit cell, normalised by the cell volume.
      In keeping with convention, a negative stress indicates the system is
      under static compression (wants to expand).
    * ``"virial"``: as ``-stress * volume``. A negative virial indicates the
      system is under static tension (wants to contract).

    For more details on how these are calculated, see :doc:`../theory`.

    :class:`~graph_pes.GraphPESModel` objects save various peices of extra
    metadata to the ``state_dict`` via the
    :meth:`~graph_pes.GraphPESModel.get_extra_state` and
    :meth:`~graph_pes.GraphPESModel.set_extra_state` methods.
    If you want to save additional extra state to the ``state_dict`` of your
    model, please implement the :meth:`~graph_pes.GraphPESModel.extra_state`
    property and corresponding setter to ensure that you do not overwrite
    these extra metadata items.

    Parameters
    ----------
    cutoff
        The cutoff radius for the model.
    implemented_properties
        The property predictions that the model implements in the forward pass.
        Must include at least ``"local_energies"``.
    """

    def __init__(
        self,
        cutoff: float,
        implemented_properties: list[PropertyKey],
    ):
        super().__init__()

        self._GRAPH_PES_VERSION: Final[str] = "0.0.28"

        self.cutoff: torch.Tensor
        self.register_buffer("cutoff", torch.tensor(cutoff))
        self._has_been_pre_fit: torch.Tensor
        self.register_buffer("_has_been_pre_fit", torch.tensor(0))

        self.implemented_properties = implemented_properties
        if "local_energies" not in implemented_properties:
            raise ValueError(
                'All GraphPESModel\'s must implement a "local_energies" '
                "prediction."
            )

    @abstractmethod
    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        """
        The model's forward pass. Generate all properties for the given graph
        that are in this model's ``implemented_properties`` list.

        Parameters
        ----------
        graph
            The graph representation of the structure/s.

        Returns
        -------
        dict[PropertyKey, torch.Tensor]
            A dictionary mapping each implemented property to a tensor of
            predictions (see above for the expected shapes). Use
            :func:`~graph_pes.atomic_graph.is_batch` to check if the
            graph is batched in the forward pass.
        """
        ...

    def predict(
        self,
        graph: AtomicGraph,
        properties: list[PropertyKey],
    ) -> dict[PropertyKey, torch.Tensor]:
        """
        Generate (optionally batched) predictions for the given
        ``properties`` and  ``graph``.

        This method returns a dictionary mapping each requested
        ``property`` to a tensor of predictions, relying on the model's
        :meth:`~graph_pes.GraphPESModel.forward` implementation
        together with :func:`torch.autograd.grad` to automatically infer any
        missing properties.

        Parameters
        ----------
        graph
            The graph representation of the structure/s.
        properties
            The properties to predict. Can be any combination of
            ``"energy"``, ``"forces"``, ``"stress"``, ``"virial"``, and
            ``"local_energies"``.
        """

        # before anything, remove unnecessary edges:
        graph = trim_edges(graph, self.cutoff.item())

        # check to see if we need to infer any properties
        infer_forces = (
            "forces" in properties
            and "forces" not in self.implemented_properties
        )
        infer_stress_information = (
            # we need to infer stress information if we're asking for
            # the stress or the virial and the model doesn't
            # implement either of them direct
            ("stress" in properties or "virial" in properties)
            and "stress" not in self.implemented_properties
            and "virial" not in self.implemented_properties
        )
        if infer_stress_information and not has_cell(graph):
            raise ValueError("Can't predict stress without cell information.")
        infer_energy = (
            any([infer_stress_information, infer_forces])
            or "energy" not in self.implemented_properties
        )

        existing_positions = graph.R
        existing_cell = graph.cell

        # inference specific set up
        if infer_stress_information:
            # See About>Theory in the graph-pes for an explanation of the
            # maths behind this.
            #
            # The stress tensor is the gradient of the total energy wrt
            # a symmetric expansion of the structure (i.e. that acts on
            # both the cell and the atomic positions).
            #
            # F. Knuth et al. All-electron formalism for total energy strain
            # derivatives and stress tensor components for numeric atom-centered
            # orbitals. Computer Physics Communications 190, 33–50 (2015).

            change_to_cell = torch.zeros_like(existing_cell)
            change_to_cell.requires_grad_(True)
            symmetric_change = 0.5 * (
                change_to_cell + change_to_cell.transpose(-1, -2)
            )  # (n_structures, 3, 3) if batched, else (3, 3)
            scaling = (
                torch.eye(3, device=existing_cell.device) + symmetric_change
            )

            # torchscript annoying-ness:
            graph_batch = graph.batch
            if graph_batch is not None:
                scaling_per_atom = torch.index_select(
                    scaling,
                    dim=0,
                    index=graph_batch,
                )  # (n_atoms, 3, 3)

                # to go from (N, 3) @ (N, 3, 3) -> (N, 3), we need un/squeeze:
                # (N, 1, 3) @ (N, 3, 3) -> (N, 1, 3) -> (N, 3)
                new_positions = (
                    graph.R.unsqueeze(-2) @ scaling_per_atom
                ).squeeze()
                # (M, 3, 3) @ (M, 3, 3) -> (M, 3, 3)
                new_cell = existing_cell @ scaling

            else:
                # (N, 3) @ (3, 3) -> (N, 3)
                new_positions = graph.R @ scaling
                new_cell = existing_cell @ scaling

            # change to positions will be a tensor of all 0's, but will allow
            # gradients to flow backwards through the energy calculation
            # and allow us to calculate the stress tensor as the gradient
            # of the energy wrt the change in cell.

            graph = replace(graph, R=new_positions, cell=new_cell)

        else:
            change_to_cell = torch.zeros_like(graph.cell)

        if infer_forces:
            graph.R.requires_grad_(True)

        # get the implemented properties
        predictions = self(graph)

        if infer_energy:
            if "local_energies" not in predictions:
                raise ValueError("Can't infer energy without local energies.")

            predictions["energy"] = sum_per_structure(
                predictions["local_energies"],
                graph,
            )

        # use the autograd machinery to auto-magically
        # calculate forces and stress from the energy

        cell_volume = torch.det(graph.cell)
        if is_batch(graph):
            cell_volume = cell_volume.view(-1, 1, 1)

        # ugly triple if loops to be efficient with autograd
        # while also Torchscript compatible
        if infer_forces and infer_stress_information:
            assert "energy" in predictions
            dE_dR, dE_dC = differentiate_all(
                predictions["energy"],
                [graph.R, change_to_cell],
                keep_graph=self.training,
            )
            predictions["forces"] = -dE_dR
            predictions["virial"] = -dE_dC
            predictions["stress"] = dE_dC / cell_volume

        elif infer_forces:
            assert "energy" in predictions
            dE_dR = differentiate(
                predictions["energy"],
                graph.R,
                keep_graph=self.training,
            )
            predictions["forces"] = -dE_dR
        elif infer_stress_information:
            assert "energy" in predictions
            dE_dC = differentiate(
                predictions["energy"],
                change_to_cell,
                keep_graph=self.training,
            )
            predictions["virial"] = -dE_dC
            predictions["stress"] = dE_dC / cell_volume

        # finally, we might not have needed autograd to infer stress/virial
        # if the other was implemented on the base class:
        if not infer_stress_information:
            if (
                "stress" in properties
                and "stress" not in self.implemented_properties
            ):
                predictions["stress"] = -predictions["virial"] / cell_volume
            if (
                "virial" in properties
                and "virial" not in self.implemented_properties
            ):
                predictions["virial"] = -predictions["stress"] * cell_volume

        # put things back to how they were before
        graph = replace(graph, R=existing_positions, cell=existing_cell)

        # make sure we don't leave auxiliary predictions
        # e.g. local_energies if we only asked for energy
        #      or energy if we only asked for forces
        predictions = {prop: predictions[prop] for prop in properties}

        # tidy up if in eval mode
        if not self.training:
            predictions = {k: v.detach() for k, v in predictions.items()}

        # return the output
        return predictions  # type: ignore

    @torch.no_grad()
    def pre_fit_all_components(
        self,
        graphs: Sequence[AtomicGraph],
    ):
        """
        Pre-fit the model, and all its components, to the training data.

        Some models require pre-fitting to the training data to set certain
        parameters. For example, the :class:`~graph_pes.models.pairwise.LennardJones`
        model uses the distribution of interatomic distances in the training data
        to set the length-scale parameter.

        In the ``graph-pes-train`` routine, this method is called before
        "normal" training begins (you can turn this off with a config option).

        This method does two things:

        1. iterates over all the model's :class:`~torch.nn.Module` components
           (inlcuding itself) and calls their :meth:`pre_fit` method (if it exists -
           see for instance :class:`~graph_pes.models.LearnableOffset` for
           an example of a model-specific pre-fit method, and
           :class:`~graph_pes.models.components.scaling.LocalEnergiesScaler` for
           an example of a component-specific pre-fit method).
        2. registers all the unique atomic numbers in the training data with
           all of the model's :class:`~graph_pes.utils.nn.PerElementParameter`
           instances to ensure correct parameter counting.

        If the model has already been pre-fitted, subsequent calls to
        :meth:`pre_fit_all_components` will be ignored (and a warning will be raised).

        Parameters
        ----------
        graphs
            The training data.
        """  # noqa: E501

        model_name = self.__class__.__name__
        logger.debug(f"Attempting to pre-fit {model_name}")

        # 1. get the graphs as a single batch
        if isinstance(graphs, GraphDataset):
            graphs = list(graphs)
        graph_batch = to_batch(graphs)

        # 2a. if the graph has already been pre-fitted: warn
        if self._has_been_pre_fit:
            model_name = self.__class__.__name__
            logger.warning(
                f"This model ({model_name}) has already been pre-fitted. "
                "This, and any subsequent, call to pre_fit_all_components will "
                "be ignored.",
                stacklevel=2,
            )

        # 2b. if the model has not been pre-fitted: pre-fit
        else:
            if len(graphs) > 10_000:
                logger.warning(
                    f"Pre-fitting on a large dataset ({len(graphs):,} graphs). "
                    "This may take some time. Consider using a smaller, "
                    "representative collection of structures for pre-fitting. "
                    "Set ``max_n_pre_fit`` in your config, or "
                    "see GraphDataset.sample() for more information.",
                    stacklevel=2,
                )

            self.pre_fit(graph_batch)
            # pre-fit any sub-module with a pre_fit method
            for module in self.modules():
                if hasattr(module, "pre_fit") and module is not self:
                    module.pre_fit(graph_batch)

            self._has_been_pre_fit = torch.tensor(1)

        # 3. finally, register all per-element parameters (no harm in doing this
        #    multiple times)
        for param in self.parameters():
            if isinstance(param, PerElementParameter):
                param.register_elements(torch.unique(graph_batch.Z).tolist())

    def pre_fit(self, graphs: AtomicGraph) -> None:
        """
        Override this method to perform additional pre-fitting steps.

        See :class:`~graph_pes.models.components.scaling.LocalEnergiesScaler` or
        :class:`~graph_pes.models.offsets.EnergyOffset` for examples of this.

        Parameters
        ----------
        graphs
            The training data.
        """

    def non_decayable_parameters(self) -> list[torch.nn.Parameter]:
        """
        Return a list of parameters that should not be decayed during training.

        By default, this method recurses over all available sub-modules
        and calls their :meth:`non_decayable_parameters` (if it is defined).

        See :class:`~graph_pes.models.components.scaling.LocalEnergiesScaler`
        for an example of this.
        """
        found = []
        for module in self.modules():
            if module is self:
                continue
            if hasattr(module, "non_decayable_parameters"):
                found.extend(module.non_decayable_parameters())
        return found

    # add type hints for mypy etc.
    def __call__(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        return super().__call__(graph)

    def get_all_PES_predictions(
        self, graph: AtomicGraph
    ) -> dict[PropertyKey, torch.Tensor]:
        """
        Get all the properties that the model can predict
        for the given ``graph``.
        """
        properties: list[PropertyKey] = [
            "energy",
            "forces",
            "local_energies",
        ]
        if has_cell(graph):
            properties.extend(["stress", "virial"])
        return self.predict(graph, properties)

    def predict_energy(self, graph: AtomicGraph) -> torch.Tensor:
        """Convenience method to predict just the energy."""
        return self.predict(graph, ["energy"])["energy"]

    def predict_forces(self, graph: AtomicGraph) -> torch.Tensor:
        """Convenience method to predict just the forces."""
        return self.predict(graph, ["forces"])["forces"]

    def predict_stress(self, graph: AtomicGraph) -> torch.Tensor:
        """Convenience method to predict just the stress."""
        return self.predict(graph, ["stress"])["stress"]

    def predict_virial(self, graph: AtomicGraph) -> torch.Tensor:
        """Convenience method to predict just the virial."""
        return self.predict(graph, ["virial"])["virial"]

    def predict_local_energies(self, graph: AtomicGraph) -> torch.Tensor:
        """Convenience method to predict just the local energies."""
        return self.predict(graph, ["local_energies"])["local_energies"]

    @torch.jit.unused
    @property
    def elements_seen(self) -> list[str]:
        """The elements that the model has seen during training."""
        Zs = set()
        for param in self.parameters():
            if isinstance(param, PerElementParameter):
                Zs.update(param._accessed_Zs)
        return [chemical_symbols[Z] for Z in sorted(Zs)]

    @torch.jit.unused
    @property
    def device(self) -> torch.device:
        return self.cutoff.device

    @torch.jit.unused
    @final
    def get_extra_state(self) -> dict[str, Any]:
        """
        Get the extra state of this instance. Please override the
        :meth:`~graph_pes.GraphPESModel.extra_state` property to add extra
        state here.
        """
        return {
            "_GRAPH_PES_VERSION": self._GRAPH_PES_VERSION,
            "extra": self.extra_state,
        }

    @torch.jit.unused
    @final
    def set_extra_state(self, state: dict[str, Any]) -> None:  # type: ignore
        """
        Set the extra state of this instance using a dictionary mapping strings
        to values returned by the :meth:`~graph_pes.GraphPESModel.extra_state`
        property setter to add extra state here.
        """
        version = state.pop("_GRAPH_PES_VERSION", None)
        if version is not None:
            current_version = self._GRAPH_PES_VERSION
            if version != current_version:
                warnings.warn(
                    "You are attempting to load a state dict corresponding "
                    f"to graph-pes version {version}, but the current version "
                    f"of this model is {current_version}. This may cause "
                    "errors when loading the model.",
                    stacklevel=2,
                )
            self._GRAPH_PES_VERSION = version  # type: ignore

        # user defined extra state
        self.extra_state = state["extra"]

    @torch.jit.unused
    @property
    def extra_state(self) -> Any:
        """
        Override this property to add extra state to the model's
        ``state_dict``.
        """
        return {}

    @torch.jit.unused
    @extra_state.setter
    def extra_state(self, state: Any) -> None:
        """
        Set the extra state of this instance using a value returned by the
        :meth:`~graph_pes.GraphPESModel.extra_state` property.
        """
        pass

    @torch.jit.unused
    def ase_calculator(
        self, device: torch.device | str | None = None, skin: float = 1.0
    ) -> "GraphPESCalculator":
        """
        Return an ASE calculator wrapping this model. See
        :class:`~graph_pes.utils.calculator.GraphPESCalculator` for more
        information.

        Parameters
        ----------
        device
            The device to use for the calculator. If ``None``, the device of the
            model will be used.
        skin
            The skin to use for the neighbour list. If all atoms have moved less
            than half of this distance between calls to `calculate`, the
            neighbour list will be reused, saving (in some cases) significant
            computation time.
        """
        from graph_pes.utils.calculator import GraphPESCalculator

        return GraphPESCalculator(self, device=device, skin=skin)



from __future__ import annotations

import warnings
from typing import Iterator, Sequence

import torch.utils.data

from ..atomic_graph import AtomicGraph, to_batch
from .datasets import GraphDataset


class GraphDataLoader(torch.utils.data.DataLoader):
    r"""
    A helper class for merging :class:`~graph_pes.AtomicGraph` objects
    into a single batch, represented as another :class:`~graph_pes.AtomicGraph`
    containing disjoint subgraphs per structure (see
    :func:`~graph_pes.atomic_graph.to_batch`).

    Parameters
    ----------
    dataset: GraphDataset | Sequence[AtomicGraph]
        The dataset to load.
    batch_size
        The batch size.
    shuffle
        Whether to shuffle the dataset.
    **kwargs:
        Additional keyword arguments to pass to the underlying
        :class:`torch.utils.data.DataLoader`.
    """

    def __init__(
        self,
        dataset: GraphDataset | Sequence[AtomicGraph],
        batch_size: int = 1,
        shuffle: bool = False,
        **kwargs,
    ):
        if not isinstance(dataset, GraphDataset):
            dataset = GraphDataset(dataset)

        if "collate_fn" in kwargs:
            warnings.warn(
                "graph-pes uses a custom collate_fn (`collate_atomic_graphs`), "
                "are you sure you want to override this?",
                stacklevel=2,
            )

        collate_fn = kwargs.pop("collate_fn", to_batch)

        super().__init__(
            dataset,
            batch_size,
            shuffle,
            collate_fn=collate_fn,
            **kwargs,
        )

    def __iter__(self) -> Iterator[AtomicGraph]:  # type: ignore
        return super().__iter__()



from __future__ import annotations

import pathlib
from typing import Sequence, overload

import ase
import ase.db
import numpy as np

from graph_pes.utils.misc import slice_to_range


class ASEDatabase(Sequence[ase.Atoms]):
    """
    A class that wraps an ASE database file, allowing for indexing into the
    database to obtain :class:`ase.Atoms` objects.

    We assume that each row contains labels in the ``data`` attribute,
    as a mapping from property names to values, and that units are "standard"
    ASE units, e.g. ``eV``, ``eV/Å``, etc.

    Fully compatible with `SchNetPack Dataset Files <https://schnetpack.readthedocs.io/en/latest/tutorials/tutorial_01_preparing_data.html>`__.

    See the `ASE documentation <https://wiki.fysik.dtu.dk/ase/ase/db/db.html>`__
    for more details about this file format.

    .. warning::

        This dataset indexes into a database, performing many random access
        reads from disk. This can be very slow! If you are using a distributed
        compute cluster, ensure you copy your database file to somewhere with
        fast local storage (as opposed to network-attached storage).

        Similarly, consider using several workers when loading the dataset,
        e.g. ``fitting/loader_kwargs/num_workers=8``.

    Parameters
    ----------
    path: str | pathlib.Path
        The path to the database.
    """  # noqa: E501

    def __init__(self, path: str | pathlib.Path):
        path = pathlib.Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Database file {path} does not exist")
        self.path = path
        self.db = ase.db.connect(path, use_lock_file=False)

    @overload
    def __getitem__(self, index: int) -> ase.Atoms: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[ase.Atoms]: ...
    def __getitem__(
        self, index: int | slice
    ) -> ase.Atoms | Sequence[ase.Atoms]:
        if isinstance(index, slice):
            indices = slice_to_range(index, len(self))
            return [self[i] for i in indices]

        atoms = self.db.get_atoms(index + 1, add_additional_information=True)
        data = atoms.info.pop("data", {})
        arrays = {
            k: v
            for k, v in data.items()
            if isinstance(v, np.ndarray) and v.shape[0] == len(atoms)
        }
        info = {k: v for k, v in data.items() if k not in arrays}
        atoms.arrays.update(arrays)
        atoms.info.update(info)
        return atoms

    def __len__(self) -> int:
        return self.db.count()



from __future__ import annotations

import os
from typing import Final

from graph_pes.utils.misc import silently_create_trainer

# dirty hack: just get lightning to work this out,
# and ensure no annoying printing happens
_trainer = silently_create_trainer(logger=False)

GLOBAL_RANK: Final[int] = _trainer.global_rank
WORLD_SIZE: Final[int] = _trainer.world_size
IS_RANK_0: Final[bool] = GLOBAL_RANK == 0


def send_to_other_ranks(key: str, value: str) -> None:
    """Must be called by rank 0 and before `Trainer.fit` is called."""
    os.environ[key] = value


def receive_from_rank_0(key: str) -> str:
    """Must be called from a non-0 rank."""
    return os.environ[key]



from __future__ import annotations

import torch
from torch import Tensor, nn

from graph_pes.atomic_graph import (
    DEFAULT_CUTOFF,
    AtomicGraph,
    PropertyKey,
    index_over_neighbours,
    neighbour_distances,
    neighbour_vectors,
    number_of_atoms,
    sum_over_neighbours,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.components.scaling import LocalEnergiesScaler
from graph_pes.utils.nn import (
    MLP,
    HaddamardProduct,
    PerElementEmbedding,
    UniformModuleList,
)

from .components.distances import Bessel, PolynomialEnvelope


class Interaction(nn.Module):
    r"""
    The interaction block of the :class:`PaiNN` model.

    Continuous filters generated from neighbour distances are convolved with
    existing scalar embeddings to create messages :math:`x_{j \rightarrow i}`
    for each neighbour :math:`j` of atom :math:`i`.

    Scalar total messages, :math:`\Delta s_i`, are created by summing over
    neighbours, while vector total messages, :math:`\Delta v_i`,
    incorporate directional information from neighbour unit vectors and
    existing vector embeddings.

    The code aims to follow **Figure 2b** of the `PaiNN paper
    <https://arxiv.org/abs/2102.03150>`_ as closely as possible.

    Parameters
    ----------
    radial_features
        The number of radial features to expand bond distances into.
    channels
        The number of channels of the internal representations.
    cutoff
        The cutoff distance for the radial features.
    """

    def __init__(
        self,
        radial_features: int,
        channels: int,
        cutoff: float,
    ):
        super().__init__()
        self.channels = channels

        self.filter_generator = HaddamardProduct(
            nn.Sequential(
                Bessel(radial_features, cutoff),
                nn.Linear(radial_features, channels * 3),
            ),
            PolynomialEnvelope(cutoff),
        )

        self.Phi = MLP(
            [channels, channels, channels * 3],
            activation=nn.SiLU(),
        )

    def forward(
        self,
        vector_embeddings: Tensor,
        scalar_embeddings: Tensor,
        graph: AtomicGraph,
    ) -> tuple[Tensor, Tensor]:
        d = neighbour_distances(graph).unsqueeze(-1)  # (E, 1)
        unit_vectors = neighbour_vectors(graph) / d  # (E, 3)

        # continous filter message creation
        new_features = self.Phi(scalar_embeddings)  # (N, 3D)
        edge_embeddings = self.filter_generator(d)  # (E, 3D)
        x_ij = index_over_neighbours(new_features, graph) * edge_embeddings
        a, b, c = torch.split(x_ij, self.channels, dim=-1)  # (E, D)

        # simple sum over neighbours to get scalar messages
        delta_s = sum_over_neighbours(a, graph)

        # create vector messages
        v_ij = b.unsqueeze(-1) * unit_vectors.unsqueeze(1)  # (E, D, 3)
        v_ij = v_ij + c.unsqueeze(-1) * index_over_neighbours(
            vector_embeddings, graph
        )
        delta_v = sum_over_neighbours(v_ij, graph)

        return delta_v, delta_s  # (N, D, 3), (N, D)


class VectorLinear(nn.Module):
    def __init__(self, in_features: int, out_features: int):
        super().__init__()
        self._linear = nn.Linear(in_features, out_features, bias=False)

    def forward(self, x: Tensor) -> Tensor:
        # a hack to swap the vector and channel dimensions
        return self._linear(x.transpose(-1, -2)).transpose(-1, -2)


class Update(nn.Module):
    r"""
    The update block of the :class:`PaiNN` model.

    Projections of vector embeddings are used to update the scalar embeddings,
    and vice versa. The code aims to follow **Figure 2c** of the `PaiNN paper
    <https://arxiv.org/abs/2102.03150>`_ as closely as possible.

    Parameters
    ----------
    channels
        The number of channels of the internal representations.
    """

    def __init__(self, channels: int):
        super().__init__()
        self.channels = channels
        self.U = VectorLinear(channels, channels)
        self.V = VectorLinear(channels, channels)
        self.mlp = MLP(
            [channels * 2, channels, channels * 3],
            activation=nn.SiLU(),
        )

    def forward(
        self, vector_embeddings: Tensor, scalar_embeddings: Tensor
    ) -> tuple[Tensor, Tensor]:
        u = self.U(vector_embeddings)  # (N, D, 3)
        v = self.V(vector_embeddings)  # (N, D, 3)

        # stack scalar message and the norm of v
        vnorm = torch.sqrt(torch.sum(v**2, dim=-1) + 1e-8)
        m = torch.cat([scalar_embeddings, vnorm], dim=-1)  # (N, 2D)
        m = self.mlp(m)  # (N, 3D)

        # split the update into 3 parts
        a, b, c = torch.split(m, self.channels, dim=-1)  # (N, D)

        # vector update:
        delta_v = u * a.unsqueeze(-1)  # (N, D, 3)

        # scalar update:
        dot = torch.sum(u * v, dim=-1)  # (N, D)
        delta_s = b + c * dot  # (N, D)

        return delta_v, delta_s


class PaiNN(GraphPESModel):
    r"""
    The `Polarizable Atom Interaction Neural Network (PaiNN)
    <https://arxiv.org/abs/2102.03150>`_ model.

    Alternating :class:`~graph_pes.models.painn.Interaction` and
    :class:`~graph_pes.models.painn.Update` blocks
    are used to residually update both vector and scalar per-atom embeddings.

    Citation:

    .. code-block:: bibtex

        @misc{Schutt-21-06,
            title = {
                Equivariant Message Passing for the Prediction
                of Tensorial Properties and Molecular Spectra
            },
            author = {
                Sch{\"u}tt, Kristof T. and Unke, Oliver T.
                and Gastegger, Michael
            },
            year = {2021},
            doi = {10.48550/arXiv.2102.03150},
        }

    Parameters
    ----------
    channels
        The number of channels of the internal representations.
    radial_features
        The number of radial features to expand bond distances into.
    layers
        The number of (interaction + update) layers to use.
    cutoff
        The cutoff distance for the radial features.

    Examples
    --------

    Configure a PaiNN model for use with ``graph-pes-train``:

    .. code:: yaml

        model:
          +PaiNN:
            channels: 32
            layers: 3
            cutoff: 5.0
    """  # noqa: E501

    def __init__(
        self,
        channels: int = 32,
        radial_features: int = 20,
        layers: int = 3,
        cutoff: float = DEFAULT_CUTOFF,
    ):
        super().__init__(
            cutoff=cutoff,
            implemented_properties=["local_energies"],
        )

        self.channels = channels
        self.z_embedding = PerElementEmbedding(channels)
        self.interactions = UniformModuleList(
            Interaction(radial_features, channels, cutoff)
            for _ in range(layers)
        )
        self.updates = UniformModuleList(
            Update(channels) for _ in range(layers)
        )
        self.read_out = MLP(
            [channels, channels, 1],
            activation=nn.SiLU(),
        )

        self.scaler = LocalEnergiesScaler()

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, Tensor]:
        # initialise embbedings:
        # - scalars as an embedding of the atomic numbers
        scalars = self.z_embedding(graph.Z)
        # - vectors as all 0s:
        vectors = torch.zeros(
            (number_of_atoms(graph), self.channels, 3), device=graph.Z.device
        )

        # iteratively interact and update the scalar and vector embeddings
        for interaction, update in zip(self.interactions, self.updates):
            dv, ds = interaction(vectors, scalars, graph)
            vectors = vectors + dv
            scalars = scalars + ds

            dv, ds = update(vectors, scalars)
            vectors = vectors + dv
            scalars = scalars + ds

        # mlp read out
        local_energies = self.read_out(scalars).squeeze()

        # scaling
        local_energies = self.scaler(local_energies, graph)
        return {"local_energies": local_energies}



from __future__ import annotations

import math
import sys
from abc import ABC, abstractmethod

import torch

from graph_pes.utils.misc import to_significant_figures


class DistanceExpansion(torch.nn.Module, ABC):
    r"""
    Abstract base class for an expansion function, :math:`\phi(r) :
    [0, r_{\text{cutoff}}] \rightarrow \mathbb{R}^{n_\text{features}}`.

    Subclasses should implement :meth:`expand`, which must also work over
    batches:

    .. math::
        \phi(r) : [0, r_{\text{cutoff}}]^{n_\text{batch} \times 1}
        \rightarrow \mathbb{R}^{n_\text{batch} \times n_\text{features}}

    Parameters
    ----------
    n_features
        The number of features to expand into.
    cutoff
        The cutoff radius.
    trainable
        Whether the expansion parameters are trainable.
    """

    def __init__(self, n_features: int, cutoff: float, trainable: bool = True):
        super().__init__()
        self.n_features = n_features
        self.register_buffer("cutoff", torch.tensor(cutoff))
        self.trainable = trainable

    @abstractmethod
    def expand(self, r: torch.Tensor) -> torch.Tensor:
        r"""
        Perform the expansion.

        Parameters
        ----------
        r : torch.Tensor
            The distances to expand. Guaranteed to have shape :math:`(..., 1)`.
        """

    def forward(self, r: torch.Tensor) -> torch.Tensor:
        """
        Call the expansion as normal in PyTorch.

        Parameters
        ----------
        r
            The distances to expand.
        """
        if r.shape[-1] != 1:
            r = r.unsqueeze(-1)
        return self.expand(r)

    # for mypy etc.
    def __call__(self, r: torch.Tensor) -> torch.Tensor:
        return super().__call__(r)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(n_features={self.n_features}, "
            f"cutoff={to_significant_figures(self.cutoff.item(), 3)}, "
            f"trainable={self.trainable})"
        )


def get_distance_expansion(
    thing: str | type[DistanceExpansion],
) -> type[DistanceExpansion]:
    """
    Get a distance expansion class by it's name.

    Parameters
    ----------
    name
        The name of the distance expansion class.

    Example
    -------
    >>> get_distance_expansion("Bessel")
    <class 'graph_pes.models.components.distances.Bessel'>
    """
    if isinstance(thing, type) and issubclass(thing, DistanceExpansion):
        return thing

    try:
        klass = getattr(sys.modules[__name__], thing)
    except AttributeError:
        raise ValueError(f"Unknown distance expansion type: {thing}") from None

    if not isinstance(klass, type) or not issubclass(klass, DistanceExpansion):
        raise ValueError(f"{thing} is not a DistanceExpansion") from None

    return klass


class Bessel(DistanceExpansion):
    r"""
    The Bessel expansion:

    .. math::
        \phi_{n}(r) = \sqrt{\frac{2}{r_{\text{cut}}}}
        \frac{\sin(n \pi \frac{r}{r_\text{cut}})}{r} \quad n
        \in [1, n_\text{features}]

    where :math:`r_\text{cut}` is the cutoff radius and :math:`n` is the order
    of the Bessel function, as introduced in `Directional Message Passing for
    Molecular Graphs <http://arxiv.org/abs/2003.03123>`_.

    .. code::

        import torch
        from graph_pes.models.components.distances import Bessel
        import matplotlib.pyplot as plt

        cutoff = 5.0
        bessel = Bessel(n_features=4, cutoff=cutoff)
        r = torch.linspace(0, cutoff, 101) # (101,)

        with torch.no_grad():
            embedding = bessel(r) # (101, 4)

        plt.plot(r / cutoff, embedding)
        plt.xlabel(r"$r / r_c$")

    .. image:: bessel.svg
        :align: center

    Parameters
    ----------
    n_features
        The number of features to expand into.
    cutoff
        The cutoff radius.
    trainable
        Whether the expansion parameters are trainable.

    Attributes
    ----------
    frequencies
        :math:`n`, the frequencies of the Bessel functions.
    """

    def __init__(self, n_features: int, cutoff: float, trainable: bool = True):
        super().__init__(n_features, cutoff, trainable)
        self.frequencies = torch.nn.Parameter(
            torch.arange(1, n_features + 1) * math.pi / cutoff,
            requires_grad=trainable,
        )
        self.pre_factor = torch.sqrt(torch.tensor(2 / cutoff))

    def expand(self, r: torch.Tensor) -> torch.Tensor:
        numerator = self.pre_factor * torch.sin(r * self.frequencies)
        # we avoid dividing by zero by replacing any zero elements with 1
        denominator = torch.where(r == 0, torch.tensor(1.0, device=r.device), r)
        return numerator / denominator


class GaussianSmearing(DistanceExpansion):
    r"""
    A Gaussian smearing expansion:

    .. math::
        \phi_{n}(r) = \exp\left(-\frac{(r - \mu_n)^2}{2\sigma^2}\right)
        \quad n \in [1, n_\text{features}]

    where :math:`\mu_n` is the center of the :math:`n`'th Gaussian
    and :math:`\sigma` is a width shared across all the Gaussians.

    .. code::

        import torch
        from graph_pes.models.components.distances import GaussianSmearing
        import matplotlib.pyplot as plt

        cutoff = 5.0
        gaussian = GaussianSmearing(n_features=4, cutoff=cutoff)
        r = torch.linspace(0, cutoff, 101)  # (101,)

        with torch.no_grad():
            embedding = gaussian(r)  # (101, 4)

        plt.plot(r / cutoff, embedding)
        plt.xlabel(r"$r / r_c$")

    .. image:: gaussian.svg
        :align: center

    Parameters
    ----------
    n_features
        The number of features to expand into.
    cutoff
        The cutoff radius.
    trainable
        Whether the expansion parameters are trainable.

    Attributes
    ----------
    centers
        :math:`\mu_n`, the centers of the Gaussians.
    coef
        :math:`\frac{1}{2\sigma^2}`, the coefficient of the exponent.
    """

    def __init__(
        self,
        n_features: int,
        cutoff: float,
        trainable: bool = True,
    ):
        super().__init__(n_features, cutoff, trainable)

        sigma = cutoff / n_features
        self.coef = torch.nn.Parameter(
            torch.tensor(-1 / (2 * sigma**2)),
            requires_grad=trainable,
        )
        self.centers = torch.nn.Parameter(
            torch.linspace(0, cutoff, n_features),
            requires_grad=trainable,
        )

    def expand(self, r: torch.Tensor) -> torch.Tensor:
        offsets = r - self.centers
        return torch.exp(self.coef * offsets**2)


class SinExpansion(DistanceExpansion):
    r"""
    A sine expansion:

    .. math::
        \phi_{n}(r) = \sin\left(\frac{n \pi r}{r_\text{cut}}\right)
        \quad n \in [1, n_\text{features}]

    where :math:`r_\text{cut}` is the cutoff radius and :math:`n` is the
    frequency of the sine function.

    .. code::

        import torch
        from graph_pes.models.components.distances import SinExpansion
        import matplotlib.pyplot as plt

        cutoff = 5.0
        sine = SinExpansion(n_features=4, cutoff=cutoff)
        r = torch.linspace(0, cutoff, 101)  # (101,)
        with torch.no_grad():
            embedding = sine(r)  # (101, 4)

        plt.plot(r / cutoff, embedding)
        plt.xlabel(r"$r / r_c$")

    .. image:: sin.svg
        :align: center

    Parameters
    ----------
    n_features
        The number of features to expand into.
    cutoff
        The cutoff radius.
    trainable
        Whether the expansion parameters are trainable.

    Attributes
    ----------
    frequencies
        :math:`n`, the frequencies of the sine functions.
    """

    def __init__(self, n_features: int, cutoff: float, trainable: bool = True):
        super().__init__(n_features, cutoff, trainable)
        self.frequencies = torch.nn.Parameter(
            torch.arange(1, n_features + 1) * math.pi / cutoff,
            requires_grad=trainable,
        )

    def expand(self, r: torch.Tensor) -> torch.Tensor:
        return torch.sin(r * self.frequencies)


class ExponentialRBF(DistanceExpansion):
    r"""
    The exponential radial basis function expansion, as introduced in
    `PhysNet: A Neural Network for Predicting Energies, Forces, Dipole
    Moments and Partial Charges <https://pubs.acs.org/doi/10.1021/acs.jctc.9b00181>`_:

    .. math::
        \phi_{n}(r) = \exp\left(-\beta_n \cdot(\exp(-r_{ij}) - \mu_n)^2 \right)
        \quad n \in [1, n_\text{features}]

    where :math:`\beta_n` and :math:`\mu_n` are the (inverse) width and
    center of the :math:`n`'th expansion, respectively.

    Following `PhysNet <https://pubs.acs.org/doi/10.1021/acs.jctc.9b00181>`_,
    :math:`\mu_n` are evenly spaced between :math:`\exp(-r_{\text{cut}})` and
    :math:`1`, and:

    .. math::
        \left( \frac{1}{\sqrt{2}\beta_n} \right)^2 =
        \frac{1 - \exp(-r_{\text{cut}})}{n_\text{features}}

    .. code::

        import torch
        from graph_pes.models.components.distances import ExponentialRBF
        import matplotlib.pyplot as plt

        cutoff = 5.0
        rbf = ExponentialRBF(n_features=10, cutoff=cutoff)
        r = torch.linspace(0, cutoff, 101)  # (101,)
        with torch.no_grad():
            embedding = rbf(r)  # (101, 10)

        plt.plot(r / cutoff, embedding)
        plt.xlabel(r"$r / r_c$")

    .. image:: erbf.svg
        :align: center

    Parameters
    ----------
    n_features
        The number of features to expand into.
    cutoff
        The cutoff radius.
    trainable
        Whether the expansion parameters are trainable.

    Attributes
    ----------
    β
        :math:`\beta_n`, the (inverse) widths of each basis.
    centers
        :math:`\mu_n`, the centers of each basis.
    """

    def __init__(self, n_features: int, cutoff: float, trainable: bool = True):
        super().__init__(n_features, cutoff, trainable)

        c = torch.exp(-torch.tensor(cutoff))
        self.beta = torch.nn.Parameter(
            torch.ones(n_features) / (2 * (1 - c) / n_features) ** 2,
            requires_grad=trainable,
        )
        self.centers = torch.nn.Parameter(
            torch.linspace(c.item(), 1, n_features),
            requires_grad=trainable,
        )

    def expand(self, r: torch.Tensor) -> torch.Tensor:
        offsets = torch.exp(-r) - self.centers
        return torch.exp(-self.beta * offsets**2)


class Envelope(torch.nn.Module):
    """
    Any envelope function, :math:`E(r)`, for smoothing potentials
    must implement a forward method that takes in a tensor of distances
    and returns a tensor of the same shape, where the values outside the
    cutoff are set to zero.
    """

    def forward(self, r: torch.Tensor) -> torch.Tensor:
        """
        Perform the envelope function.

        Parameters
        ----------
        r : torch.Tensor
            The distances to envelope.
        """
        ...

    def __call__(self, r: torch.Tensor) -> torch.Tensor:
        return super().__call__(r)


class PolynomialEnvelope(Envelope):
    r"""
    A thrice differentiable envelope function.

    .. math::
        E_p(r) = 1 - \frac{(p+1)(p+2)}{2}\cdot r^p + p(p+2) \cdot r^{p+1}
        - \frac{p(p+1)}{2}\cdot d^{p+2}

    where :math:`r_\text{cut}` is the cutoff radius,
    and :math:`p \in \mathbb{N}`.

    Parameters
    ----------
    cutoff : float
        The cutoff radius.
    p: int
        The order of the envelope function.
    """

    def __init__(self, cutoff: float, p: int = 6):
        super().__init__()
        self.cutoff = cutoff
        self.p = p
        self.register_buffer(
            "coefficients",
            torch.tensor(
                [
                    -(p + 1) * (p + 2) / 2,
                    p * (p + 2),
                    -(p * (p + 1)) / 2,
                ]
            ),
        )
        self.register_buffer("powers", torch.arange(p, p + 3))

    def forward(self, r: torch.Tensor) -> torch.Tensor:
        powers = (r.unsqueeze(-1) / self.cutoff) ** self.powers
        envelope = 1 + (powers * self.coefficients).sum(dim=-1)
        return torch.where(
            r <= self.cutoff, envelope, torch.tensor(0.0, device=r.device)
        )

    def __repr__(self):
        return f"PolynomialEnvelope(cutoff={self.cutoff}, p={self.p})"


class CosineEnvelope(Envelope):
    r"""
    A cosine envelope function.

    .. math::
        E_c(r) = \frac{1}{2}\left(1 + \cos\left(\frac{\pi r}{r_\text{cut}}
        \right)\right)

    where :math:`r_\text{cut}` is the cutoff radius.

    Parameters
    ----------
    cutoff : float
        The cutoff radius.
    """

    def __init__(self, cutoff: float):
        super().__init__()
        self.cutoff = cutoff

    def forward(self, r: torch.Tensor) -> torch.Tensor:
        cos = 0.5 * (1 + torch.cos(math.pi * r / self.cutoff))
        return torch.where(r <= self.cutoff, cos, torch.tensor(0.0))

    def __repr__(self):
        return f"CosineEnvelope(cutoff={self.cutoff})"


class SmoothOnsetEnvelope(Envelope):
    r"""
    A smooth cutoff function with an onset.

    .. math::

        f(r, r_o, r_c) = \begin{cases}
        \hfill 1 \hfill & \text{if } r < r_o \\
        \frac{(r_c - r)^2 (r_c + 2r - 3r_o)}{(r_c - r_o)^3} & \text{if } r_o \leq r < r_c \\
        \hfill 0 \hfill & \text{if } r \geq r_c
        \end{cases}

    where :math:`r_o` is the onset radius and :math:`r_c` is the cutoff radius.

    Parameters
    ----------
    cutoff : float
        The cutoff radius.
    onset : float
        The onset radius.
    """  # noqa: E501

    def __init__(self, cutoff: float, onset: float):
        super().__init__()
        if onset >= cutoff:
            raise ValueError("Onset must be less than cutoff")

        self.register_buffer("cutoff", torch.tensor(cutoff))
        self.register_buffer("onset", torch.tensor(onset))

    def forward(self, r: torch.Tensor) -> torch.Tensor:
        return torch.where(
            r < self.onset,
            torch.tensor(1.0, device=r.device),
            torch.where(
                r < self.cutoff,
                (self.cutoff - r) ** 2
                * (self.cutoff + 2 * r - 3 * self.onset)
                / (self.cutoff - self.onset) ** 3,
                torch.tensor(0.0, device=r.device),
            ),
        )

    def __repr__(self):
        return f"SmoothOnsetEnvelope(cutoff={self.cutoff}, onset={self.onset})"



from __future__ import annotations

import copy
import logging
import random
import string
import sys
from pathlib import Path
from typing import Any, Iterable, Iterator, Sequence, TypeVar, overload

import pytorch_lightning as pl
import torch
from torch import Tensor

T = TypeVar("T")


MAX_Z = 118
"""The maximum atomic number in the periodic table."""


@overload
def pairs(a: Sequence[T]) -> Iterator[tuple[T, T]]: ...
@overload
def pairs(a: Tensor) -> Iterator[tuple[Tensor, Tensor]]: ...
def pairs(a) -> Iterator[tuple[T, T] | tuple[Tensor, Tensor]]:
    """
    Iterate over pairs of elements in `a`

    Parameters
    ----------
    a
        The sequence or tensor to iterate over.

    Example
    -------
    >>> list(pairs([1, 2, 3]))
    [(1, 2), (2, 3)]

    >>> list(pairs(Tensor([1, 2, 3])))
    [(1, 2), (2, 3)]
    """
    return zip(a, a[1:])


def as_possible_tensor(value: object) -> Tensor | None:
    """
    Convert a value to a tensor if possible.

    Parameters
    ----------
    value
        The value to convert.
    """

    try:
        tensor = torch.as_tensor(value)
        if tensor.dtype == torch.float64:
            tensor = tensor.float()
        return tensor

    except Exception:
        return None


def differentiate_all(
    y: torch.Tensor,
    xs: list[torch.Tensor],
    keep_graph: bool = False,
):
    """
    A ``Torchscript``-compatible way to differentiate `y` with respect
    to all of the tensors in `xs`, with a slightly nicer API than
    ``torch.autograd.grad``, and handling the (odd) cases where either
    or both of `y` or `xs` do not have a gradient function.
    """

    if not torch.is_grad_enabled():
        raise RuntimeError(
            "Autograd is disabled, but you are trying to "
            "calculate gradients. Please wrap your code in "
            "a torch.enable_grad() context."
        )

    defaults = [torch.zeros_like(x) for x in xs]

    x_did_require_grad = [x.requires_grad for x in xs]
    for x in xs:
        x.requires_grad_(True)

    y_total = y.sum()
    # ensure y has a grad_fn
    y_total = y_total + torch.tensor(0.0, requires_grad=True)

    grads = torch.autograd.grad(
        [y_total],
        xs,
        create_graph=keep_graph,
        allow_unused=True,
    )

    for x, did_require_grad in zip(xs, x_did_require_grad):
        x.requires_grad_(did_require_grad)

    return [
        grad if grad is not None else default
        for grad, default in zip(grads, defaults)
    ]


def differentiate(y: torch.Tensor, x: torch.Tensor, keep_graph: bool = False):
    """
    A torchscript-compatible way to differentiate `y` with respect
    to `x`, handling the (odd) cases where either or both of
    `y` or `x` do not have a gradient function: in these cases,
    we return a tensor of zeros with the correct shape and
    requires_grad set to True.
    """

    return differentiate_all(y, [x], keep_graph)[0]


def to_significant_figures(x: float | int, sf: int = 3) -> float:
    """
    Get a string representation of a float, rounded to
    `sf` significant figures.
    """

    # do the actual rounding
    possibly_scientific = f"{x:.{sf}g}"

    # this might be in e.g. 1.23e+02 format,
    # so convert to float and back to string
    return float(possibly_scientific)


def is_being_documented():
    return "sphinx" in sys.modules


def uniform_repr(
    thing_name: str,
    *anonymous_things: Any,
    max_width: int = 60,
    stringify: bool = True,
    indent_width: int = 2,
    **named_things: Any,
) -> str:
    def _to_str(thing: Any) -> str:
        if isinstance(thing, str) and stringify:
            return f'"{thing}"'
        return str(thing)

    info = list(map(_to_str, anonymous_things))
    info += [f"{name}={_to_str(thing)}" for name, thing in named_things.items()]

    single_liner = f"{thing_name}({', '.join(info)})"
    if len(single_liner) < max_width and "\n" not in single_liner:
        return single_liner

    def indent(s: str) -> str:
        _indent = " " * indent_width
        return "\n".join(f"{_indent}{line}" for line in s.split("\n"))

    # if we're here, we need to do a multi-line repr
    rep = f"{thing_name}("
    for thing in info:
        rep += "\n" + indent(thing) + ","

    # remove trailing comma, add final newline and close bracket
    return rep[:-1] + "\n)"


def nested_merge_all(*dicts: dict) -> dict:
    """
    Merge multiple nested dictionaries, with later dictionaries
    taking precedence over earlier ones.
    """

    result = {}
    for d in dicts:
        result = nested_merge(result, d)
    return result


def nested_merge(a: dict, b: dict):
    """
    Merge two nested dictionaries, with `b` taking precedence
    over `a`.
    """

    new_dict = copy.deepcopy(a)
    for key, value in b.items():
        if (
            key in new_dict
            and isinstance(value, dict)
            and isinstance(new_dict[key], dict)
        ):
            new_dict[key] = nested_merge(new_dict[key], value)
        else:
            new_dict[key] = value

    return new_dict


def build_single_nested_dict(keys: list[str], value: Any) -> dict:
    """
    Build a single nested dictionary from a list of keys and a value.
    """

    result = value
    for key in reversed(keys):
        result = {key: result}
    return result


def random_id(
    lengths: list[int] | None = None,
    use_existing: bool = False,
) -> str:
    """
    Generate a random ID of the form ``abdc123_efgh456_...``.

    Parameters
    ----------
    lengths
        The lengths of the individual parts of the ID.
    use_existing
        Whether to use the existing random number generator in ``random``,
        or to create a new one.

    Example
    -------
    >>> random_id(lengths=[4, 4])
    "abcd_1234"
    """

    if lengths is None:
        lengths = [6, 6]

    if use_existing:
        rng = random
    else:
        # seed with the current time
        rng = random.Random()
        rng.seed()

    return "_".join(
        "".join(rng.choices(string.ascii_lowercase + string.digits, k=k))
        for k in lengths
    )


def random_dir(root: Path) -> Path:
    """Find a random directory that doesn't exist in `root`."""

    while True:
        new_dir = root / random_id()
        if not new_dir.exists():
            return new_dir


def contains_tensor(l: Iterable[torch.Tensor], tensor: torch.Tensor) -> bool:
    """
    A convenient way to check if a list contains a particular tensor,
    since ``tensor in l`` is broadcasted by torch to return a boolean
    tensor with the same shape as ``l``.
    """
    return any(tensor is t for t in l)


def left_aligned_mul(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    r"""
    Calculate :math:`z = x \odot y` such that:

    .. math::

            z_{i, j, \dots} = x_{i, j, \dots} \cdot y_i

    That is, broadcast :math:`y` to the far left of :math:`x` (the opposite
    sense of normal broadcasting in torch), and multiply the two tensors
    elementwise.

    Parameters
    ----------
    x
        of shape (n, ...)
    y
        of shape (n, )

    Returns
    -------
    torch.Tensor
        of same shape as x
    """
    if x.dim() == 1 or x.dim() == 0:
        return x * y

    # x of shape (n, ..., a)
    x = x.transpose(0, -1)  # shape: (a, ..., n)
    result = x * y  # shape: (a, ..., n)
    return result.transpose(0, -1)  # shape: (n, ..., a)


def left_aligned_div(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    r"""
    Calculate :math:`z = x \oslash y` such that:

    .. math::

            z_{i, j, \dots} = x_{i, j, \dots} / y_i

    That is, broadcast :math:`y` to the far left of :math:`x` (the opposite
    sense of normal broadcasting in torch), and divide the two tensors
    elementwise.

    Parameters
    ----------
    x
        of shape (n, ...)
    y
        of shape (n, )

    Returns
    -------
    torch.Tensor
        of same shape as x
    """
    if x.dim() == 1 or x.dim() == 0:
        return x / y

    # x of shape (n, ..., a)
    x = x.transpose(0, -1)  # shape: (a, ..., n)
    result = x / y  # shape: (a, ..., n)
    return result.transpose(0, -1)  # shape: (n, ..., a)


def all_equal(iterable: Iterable[T]) -> bool:
    """
    Check if all elements in an iterable are the same. If the
    iterable is empty, return `False`.

    Returns
    -------
    bool
        Whether all elements in `iterable` are the same.
    """
    iterator = iter(iterable)
    try:
        first = next(iterator)
    except StopIteration:
        return False
    return all(first == x for x in iterator)


def groups_of(
    size: int,
    things: Iterable[T],
    drop_last: bool = False,
) -> Iterator[list[T]]:
    """
    Split an iterable into groups of a specified size.

    Parameters
    ----------
    size : int
        The size of each group.
    things : Iterable[T]
        The iterable to split into groups.
    drop_last : bool, default=False
        If True, drop the last group if it's smaller than the specified size.
        If False, yield the last group even if incomplete.

    Returns
    -------
    Iterator[list[T]]
        An iterator yielding lists of items, where each list has length `size`
        (except possibly the last one if `drop_last=False`).

    Examples
    --------
    >>> for group in groups_of(5, range(12)):
    ...     print(group)
    [0, 1, 2, 3, 4]
    [5, 6, 7, 8, 9]
    [10, 11]

    >>> list(groups_of(2, range(5), drop_last=True))
    [[0, 1], [2, 3]]
    """
    batch = []
    for thing in things:
        batch.append(thing)
        if len(batch) == size:
            yield batch
            batch = []

    if batch and not drop_last:
        yield batch


def full_3x3_to_voigt_6(tensor: torch.Tensor) -> torch.Tensor:
    """
    Convert a (optionally batched) 3x3 tensor to Voigt notation.

    Parameters
    ----------
    tensor
        A (optionally batched) 3x3 tensor, shape ``(..., 3, 3)``.

    Returns
    -------
    torch.Tensor
        A tensor of shape ``(..., 6)``, in Voigt notation.
    """
    if tensor.ndim == 2:
        tensor = tensor.unsqueeze(0)
    N = tensor.shape[0]

    voigt_6 = torch.zeros(N, 6, device=tensor.device, dtype=tensor.dtype)
    voigt_6[..., 0] = tensor[..., 0, 0]
    voigt_6[..., 1] = tensor[..., 1, 1]
    voigt_6[..., 2] = tensor[..., 2, 2]
    voigt_6[..., 3] = tensor[..., 0, 1]
    voigt_6[..., 4] = tensor[..., 0, 2]
    voigt_6[..., 5] = tensor[..., 1, 2]
    return voigt_6.squeeze()


def voigt_6_to_full_3x3(tensor: torch.Tensor) -> torch.Tensor:
    """
    Convert a (optionally batched) 6-element tensor in Voigt notation to a
    3x3 tensor.

    Parameters
    ----------
    tensor
        A (optionally batched) 6-element tensor in Voigt notation,
        shape ``(..., 6)``.

    Returns
    -------
    torch.Tensor
        A tensor of shape ``(..., 3, 3)``.
    """
    if tensor.ndim == 1:
        tensor = tensor.unsqueeze(0)
    N = tensor.shape[0]

    full_3x3 = torch.zeros(N, 3, 3, device=tensor.device, dtype=tensor.dtype)
    full_3x3[..., 0, 0] = tensor[..., 0]
    full_3x3[..., 1, 1] = tensor[..., 1]
    full_3x3[..., 2, 2] = tensor[..., 2]
    full_3x3[..., 0, 1] = tensor[..., 3]
    full_3x3[..., 1, 0] = tensor[..., 3]
    full_3x3[..., 0, 2] = tensor[..., 4]
    full_3x3[..., 2, 0] = tensor[..., 4]
    full_3x3[..., 1, 2] = tensor[..., 5]
    full_3x3[..., 2, 1] = tensor[..., 5]
    return full_3x3.squeeze()


def slice_to_range(slice_obj: slice, sequence_length: int) -> range:
    """
    Convert a slice object to a range object.

    Parameters
    ----------
    slice_obj : slice
        The slice to convert
    sequence_length : int
        The length of the sequence being sliced

    Returns
    -------
    range
        A range object representing the indices that would be selected by the
        slice.
    """
    start = 0 if slice_obj.start is None else slice_obj.start
    stop = sequence_length if slice_obj.stop is None else slice_obj.stop
    step = 1 if slice_obj.step is None else slice_obj.step

    return range(start, stop, step)


def silently_create_trainer(**trainer_kwargs) -> pl.Trainer:
    logger = logging.getLogger("pytorch_lightning.utilities.rank_zero")
    logger.setLevel(logging.ERROR)
    _trainer = pl.Trainer(**trainer_kwargs)
    logger.setLevel(logging.INFO)
    return _trainer


def first(iterable: Iterable[T]) -> T | None:
    """
    Get the first element of an iterable, or `None` if the iterable is empty.
    """
    try:
        return next(iter(iterable))
    except StopIteration:
        return None


class MultiSequence(Sequence[T]):
    """
    A class that mimics a concatenation of sequences without performing
    costly indexing operations.
    """

    def __init__(
        self,
        sequences: Sequence[Sequence[T]],
        indices: Sequence[int] | None = None,
    ):
        self.sequences = sequences
        self.indices = indices or range(sum(len(s) for s in sequences))

    def __len__(self) -> int:
        return len(self.indices)

    @overload
    def __getitem__(self, index: int) -> T: ...
    @overload
    def __getitem__(self, index: slice) -> MultiSequence[T]: ...
    def __getitem__(self, index: int | slice) -> T | MultiSequence[T]:
        """
        Get the element/s at the given ``index``.
        """
        if isinstance(index, int):
            actual_index = self.indices[index]

            for seq in self.sequences:
                if actual_index < len(seq):
                    return seq[actual_index]
                actual_index -= len(seq)
            raise IndexError(f"Index {index} is out of bounds for the sequence")

        elif isinstance(index, slice):
            return MultiSequence(self.sequences, self.indices[index])

        raise NotImplementedError(f"Index {index} is not supported")



from __future__ import annotations

import re
import warnings
from typing import TypeVar

warnings.filterwarnings(
    "ignore",
    module="e3nn",
    message=".*you are using `torch.load`.*",
)

import pathlib

import torch

from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.utils.logger import logger

from .addition import AdditionModel
from .e3nn.mace import MACE, ZEmbeddingMACE
from .e3nn.nequip import NequIP, ZEmbeddingNequIP
from .offsets import FixedOffset, LearnableOffset
from .painn import PaiNN
from .pairwise import (
    LennardJones,
    LennardJonesMixture,
    Morse,
    PairPotential,
    SmoothedPairPotential,
    ZBLCoreRepulsion,
)
from .schnet import SchNet
from .tensornet import TensorNet

__all__ = [
    "AdditionModel",
    "FixedOffset",
    "LearnableOffset",
    "LennardJones",
    "LennardJonesMixture",
    "MACE",
    "Morse",
    "NequIP",
    "PaiNN",
    "PairPotential",
    "SchNet",
    "SmoothedPairPotential",
    "TensorNet",
    "ZBLCoreRepulsion",
    "ZEmbeddingMACE",
    "ZEmbeddingNequIP",
]

MODEL_EXCLUSIONS = {
    "FixedOffset",
    "LearnableOffset",
    "AdditionModel",
    "PairPotential",
    "SmoothedPairPotential",
}

ALL_MODELS: list[type[GraphPESModel]] = [
    globals()[model] for model in __all__ if model not in MODEL_EXCLUSIONS
]


def load_model(path: str | pathlib.Path) -> GraphPESModel:
    """
    Load a model from a file.

    Parameters
    ----------
    path
        The path to the file.

    Returns
    -------
    GraphPESModel
        The model.

    Examples
    --------

    Use this function to load an existing model for further training using
    ``graph-pes-train``:

    .. code-block:: yaml

        model:
            +load_model:
                path: path/to/model.pt

    To account for some new energy offset in your training data, you could do
    something like this:
    (see also :func:`~graph_pes.models.load_model_component`)

    .. code-block:: yaml

        model:
            # add an offset to an existing model before fine-tuning
            offset: +LearnableOffset()
            many-body:
                +load_model:
                    path: path/to/model.pt
    """
    path = pathlib.Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Could not find model at {path}")

    model = torch.load(path, weights_only=False)

    if not isinstance(model, GraphPESModel):
        raise ValueError(
            "Expected the loaded object to be a GraphPESModel "
            f"but got {type(model)}"
        )

    import graph_pes

    if model._GRAPH_PES_VERSION != graph_pes.__version__:
        warnings.warn(
            "You are attempting to load a model that was trained with "
            f"a different version of graph-pes ({model._GRAPH_PES_VERSION}) "
            f"than what you are currently using ({graph_pes.__version__}). "
            "We won't stop you from doing this, but it may cause issues.",
            stacklevel=2,
        )

    return model


def load_model_component(
    path: str | pathlib.Path,
    key: str,
) -> GraphPESModel:
    """
    Load a component from an :class:`~graph_pes.models.AdditionModel`.

    Parameters
    ----------
    path
        The path to the file.
    key
        The key to load.

    Returns
    -------
    GraphPESModel
        The component.

    Examples
    --------

    Train on data with a new energy offset:

    .. code-block:: yaml

        model:
            offset: +LearnableOffset()
            many-body:
                +load_model_component:
                    path: path/to/model.pt
                    key: many-body
    """

    base_model = load_model(path)
    if not isinstance(base_model, AdditionModel):
        raise ValueError(
            f"Expected to load an AdditionModel, got {type(base_model)}"
        )

    return base_model[key]


T = TypeVar("T", bound=torch.nn.Module)


def freeze(model: T) -> T:
    """
    Freeze all parameters in a module.

    Parameters
    ----------
    model
        The model to freeze.

    Returns
    -------
    T
        The model.
    """
    for param in model.parameters():
        param.requires_grad = False
    return model


def freeze_matching(model: T, pattern: str) -> T:
    r"""
    Freeze all parameters that match the given pattern.

    Parameters
    ----------
    model
        The model to freeze.
    pattern
        The regular expression to match the names of the parameters to freeze.

    Returns
    -------
    T
        The model.

    Examples
    --------

    Freeze all the parameters in the first layer of a MACE-MP0 model from
    :func:`~graph_pes.interfaces.mace_mp` (which have names of the form
    ``"model.interactions.0.<name>"``):

    .. code-block:: yaml

        model:
            +freeze_any_matching:
                model: +mace_mp()
                pattern: model\.interactions\.0\..*
    """
    for name, param in model.named_parameters():
        if re.match(pattern, name):
            logger.info(f"Freezing {name}")
            param.requires_grad = False
    return model


def freeze_any_matching(model: T, patterns: list[str]) -> T:
    r"""
    Freeze all parameters that match any of the given patterns.

    Parameters
    ----------
    model
        The model to freeze.
    patterns
        The patterns to match.

    Returns
    -------
    T
        The model.


    """
    for pattern in patterns:
        freeze_matching(model, pattern)
    return model


def freeze_all_except(model: T, pattern: str | list[str]) -> T:
    r"""
    Freeze all parameters in a model except those matching a given pattern.

    Parameters
    ----------
    model
        The model to freeze.
    pattern
        The pattern/s to match.

    Returns
    -------
    T
        The model.

    Examples
    --------

    Freeze all parameters in a MACE-MP0 model from
    :func:`~graph_pes.interfaces.mace_mp` except those in the
    read-out heads:

    .. code-block:: yaml

        model:
            +freeze_all_except:
                model: +mace_mp()
                pattern: model\.readouts.*
    """
    freeze(model)

    if isinstance(pattern, str):
        pattern = [pattern]
    for name, param in model.named_parameters():
        if any(re.match(p, name) for p in pattern):
            param.requires_grad = True

    return model



from __future__ import annotations

import torch
from torch import Tensor, nn

from graph_pes.atomic_graph import (
    DEFAULT_CUTOFF,
    AtomicGraph,
    PropertyKey,
    index_over_neighbours,
    neighbour_distances,
    neighbour_vectors,
    number_of_edges,
    sum_over_neighbours,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.components.scaling import LocalEnergiesScaler
from graph_pes.utils.nn import (
    MLP,
    HaddamardProduct,
    PerElementEmbedding,
    UniformModuleList,
)

from .components.distances import (
    CosineEnvelope,
    DistanceExpansion,
    ExponentialRBF,
    get_distance_expansion,
)


class TensorNet(GraphPESModel):
    r"""
    The `TensorNet <http://arxiv.org/abs/2306.06482>`_ architecture.

    Citation:

    .. code:: bibtex

        @misc{Simeon-23-06,
            title = {
                TensorNet: Cartesian Tensor Representations for
                Efficient Learning of Molecular Potentials
            },
            author = {Simeon, Guillem and {de Fabritiis}, Gianni},
            year = {2023},
            number = {arXiv:2306.06482},
        }

    Parameters
    ----------
    cutoff
        The cutoff radius to use for the model.
    radial_features
        The number of radial features to use for the model.
    radial_expansion
        The type of radial basis function to use for the model.
        For more examples, see
        :class:`~graph_pes.models.components.distances.DistanceExpansion`.
    channels
        The size of the embedding for each atom.
    layers
        The number of interaction layers to use for the model.
    direct_force_predictions
        Whether to predict forces directly. If ``True``, the model will
        generate force predictions by passing the final
        layer's node embeddings through a
        :class:`~graph_pes.models.tensornet.VectorOutput` read out.
        Otherwise, ``graph-pes`` automatically infers the forces as the
        derivative of the energy with respect to the atomic positions.

    Examples
    --------

    Configure a TensorNet model for use with ``graph-pes-train``:

    .. code:: yaml

        model:
          +TensorNet:
            radial_features: 8
            radial_expansion: Bessel
            channels: 32
            cutoff: 5.0
    """

    def __init__(
        self,
        cutoff: float = DEFAULT_CUTOFF,
        radial_features: int = 32,
        radial_expansion: str | type[DistanceExpansion] = ExponentialRBF,
        channels: int = 32,
        layers: int = 2,
        direct_force_predictions: bool = False,
    ):
        properties: list[PropertyKey] = ["local_energies"]
        if direct_force_predictions:
            properties.append("forces")

        super().__init__(
            cutoff=cutoff,
            implemented_properties=properties,
        )

        self.embedding = Embedding(
            radial_features, radial_expansion, channels, cutoff
        )

        self.interactions = UniformModuleList(
            Interaction(radial_features, channels, cutoff)
            for _ in range(layers)
        )

        self.energy_read_out = ScalarOutput(channels)
        if direct_force_predictions:
            self.force_read_out = VectorOutput(channels)
        else:
            self.force_read_out = None

        self.scaler = LocalEnergiesScaler()

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        X = self.embedding(graph)  # (N, C, 3, 3)

        for interaction in self.interactions:
            # normalise -> interaction -> residual connection
            X = X / (frobenius_norm(X)[..., None, None] + 1)
            dX = interaction(X, graph)
            X = X + dX

        local_energies = self.energy_read_out(X).squeeze()
        local_energies = self.scaler(local_energies, graph)

        results: dict[PropertyKey, torch.Tensor] = {
            "local_energies": local_energies
        }

        if self.force_read_out is not None:
            results["forces"] = self.force_read_out(X)

        return results


### components ###


class EdgeEmbedding(nn.Module):
    r"""
    Generates embeddings for each (directed) edge in the graph, incorporating
    the species of each atom, and the vector between them.

    1. generate initial edge embedding components:

       * :math:`I_0^{(ij)} = \text{Id}`
       * :math:`A_0^{(ij)} = \begin{pmatrix} 
         0 & \hat{r}_{ij}^z & - \hat{r}_{ij}^y \\
         - \hat{r}_{ij}^z & 0 & \hat{r}_{ij}^x \\
         \hat{r}_{ij}^y & - \hat{r}_{ij}^x & 0
         \end{pmatrix}`
       * :math:`S_0^{(ij)} = \hat{r}_{ij} \cdot \hat{r}_{ij}^T -
         \text{Tr}(\hat{r}_{ij} \cdot \hat{r}_{ij}^T) \text{Id}`

    2. generate an embedding of the ordered pair of species:

    .. math::
    
       h_Z^{(ij)} = f(z_i, z_j) = \text{Linear}(\text{embed}(z_i) 
       || \text{embed}(z_j))

    3. expand the edge vectors in to an exponential radial basis:

    .. math::
      
         h_{R,n}^{(ij)} = \exp(\beta_n \cdot (\exp(- r_{ij}) - \mu_k)^2)

    4. combine all edge embeddings:

    .. math::

        X^{ij} = \phi(r_{ij}) \cdot h_Z^{(ij)} \cdot \left ( 
        \text{Dense}(h_R^{(ij)}) \cdot I_0^{(ij)} +
        \text{Dense}(h_R^{(ij)}) \cdot A_0^{(ij)} +
        \text{Dense}(h_R^{(ij)}) \cdot S_0^{(ij)}
        \right)

    where :math:`\phi(r_{ij})` is the cosine envelope function.
    """

    def __init__(
        self,
        radial_features: int,
        radial_expansion: str | type[DistanceExpansion],
        channels: int,
        cutoff: float,
    ):
        super().__init__()

        self.z_embedding = PerElementEmbedding(channels)
        self.z_map = nn.Linear(2 * channels, channels, bias=False)
        expansion_klass = get_distance_expansion(radial_expansion)
        self.distance_embedding = HaddamardProduct(
            nn.Sequential(
                expansion_klass(radial_features, cutoff),
                nn.Linear(radial_features, 3 * channels),
            ),
            CosineEnvelope(cutoff),
            left_aligned=True,
        )

    def forward(self, graph: AtomicGraph) -> tuple[Tensor, Tensor, Tensor]:
        C = self.z_embedding.dim()
        E = number_of_edges(graph)

        # 1. generate initial edge embedding components:
        I_0, A_0, S_0 = self._initial_edge_embeddings(graph)  # (E, 1, 3, 3)

        # 2. encode atomic species of ordered neighbour pairs:
        h_z_atom = self.z_embedding(graph.Z)  # (N, C)
        h_z_edge = h_z_atom[graph.neighbour_list]  # (2, E, C)
        h_z_edge = h_z_edge.permute(1, 0, 2).reshape(E, 2 * C)
        h_z_edge = self.z_map(h_z_edge)  # (E, C)

        # 3. embed edge distances
        h_r = self.distance_embedding(neighbour_distances(graph))  # (E, 3C)

        # 4. combine information into coefficients
        c = (h_r * h_z_edge.repeat(1, 3))[..., None, None]  # (E, 3C, 1, 1)
        c_I, c_A, c_S = torch.chunk(c, 3, dim=1)  # (E, C, 1, 1)

        return c_I * I_0, c_A * A_0, c_S * S_0  # 3x (E, C, 3, 3)

    def _initial_edge_embeddings(
        self, graph: AtomicGraph
    ) -> tuple[Tensor, Tensor, Tensor]:
        E = number_of_edges(graph)
        r_hat = neighbour_vectors(graph) / neighbour_distances(graph)[..., None]
        eye = torch.eye(3, device=graph.Z.device)
        I_ij = torch.repeat_interleave(eye[None, ...], E, dim=0)  # (E, 3, 3)
        A_ij = vector_to_skew_symmetric_matrix(r_hat)
        S_ij = vector_to_symmetric_traceless_matrix(r_hat)
        return (
            I_ij.view(E, 1, 3, 3),
            A_ij.view(E, 1, 3, 3),
            S_ij.view(E, 1, 3, 3),
        )


class Embedding(nn.Module):
    """
    Embed the local environment of each atom into a ``(C, 3, 3)`` tensor.
    """

    def __init__(
        self,
        radial_features: int,
        radial_expansion: str | type[DistanceExpansion],
        channels: int,
        cutoff: float,
    ):
        super().__init__()
        self.edge_embedding = EdgeEmbedding(
            radial_features, radial_expansion, channels, cutoff
        )
        self.layer_norm = nn.LayerNorm(channels)
        self.mlp = MLP(
            layers=[
                channels,
                2 * channels,
                3 * channels,
            ],
            activation=nn.SiLU(),
            activate_last=True,
        )
        self.W_I = TensorLinear(channels, channels)
        self.W_A = TensorLinear(channels, channels)
        self.W_S = TensorLinear(channels, channels)

    def forward(self, graph: AtomicGraph) -> Tensor:
        # embed edges
        I_edge, A_edge, S_edge = self.edge_embedding(graph)  # (E, C, 3, 3)

        # sum over neighbours to get atom embeddings
        I_atom = sum_over_neighbours(I_edge, graph)  # (N, C, 3, 3)
        A_atom = sum_over_neighbours(A_edge, graph)  # (N, C, 3, 3)
        S_atom = sum_over_neighbours(S_edge, graph)  # (N, C, 3, 3)

        # generate coefficients from tensor representations
        # (mixes irreps)
        norms = frobenius_norm(I_atom + A_atom + S_atom)  # (N, C)
        norms = self.layer_norm(norms)
        coefficients = self.mlp(norms)[..., None, None]  # (N, 3C, 1, 1)
        c_I, c_A, c_S = torch.chunk(coefficients, 3, dim=1)  # (N, C, 1, 1)

        # ...and combine with mixed coefficients with linear
        # mixing of features
        return (
            c_I * self.W_I(I_atom)
            + c_A * self.W_A(A_atom)
            + c_S * self.W_S(S_atom)
        )  # (N, C, 3, 3)


class Interaction(nn.Module):
    def __init__(
        self,
        radial_features: int,
        channels: int,
        cutoff: float,
    ):
        super().__init__()

        # unfortunately, we need to be explicit to satisfy torchscript
        self.W_I_pre = TensorLinear(channels, channels)
        self.W_A_pre = TensorLinear(channels, channels)
        self.W_S_pre = TensorLinear(channels, channels)
        self.W_I_post = TensorLinear(channels, channels)
        self.W_A_post = TensorLinear(channels, channels)
        self.W_S_post = TensorLinear(channels, channels)

        self.distance_embedding = HaddamardProduct(
            nn.Sequential(
                ExponentialRBF(radial_features, cutoff),
                MLP(
                    layers=[radial_features, 2 * channels, 3 * channels],
                    activation=nn.SiLU(),
                    activate_last=True,
                ),
            ),
            CosineEnvelope(cutoff),
            left_aligned=True,
        )

    def forward(self, X: Tensor, graph: AtomicGraph) -> Tensor:
        # decompose matrix representations
        I, A, S = decompose_matrix(X)  # (N, C, 3, 3)

        # update I, A, S
        Y = self.W_I_pre(I) + self.W_A_pre(A) + self.W_S_pre(S)  # (N, C, 3, 3)

        # get coefficients from neighbour distances
        c = self.distance_embedding(neighbour_distances(graph))  # (E, 3C)
        f_I, f_A, f_S = torch.chunk(c[..., None, None], 3, dim=1)

        # message passing
        m_ij = (
            f_I * index_over_neighbours(I, graph)
            + f_A * index_over_neighbours(A, graph)
            + f_S * index_over_neighbours(S, graph)
        )
        # total message
        M_i = sum_over_neighbours(m_ij, graph)  # (N, C, 3, 3)

        # scalar/vector/tensor mixing
        Y = Y @ M_i + M_i @ Y  # (N, C, 3, 3)

        # renormalise and decompose
        norm = (frobenius_norm(Y) + 1)[..., None, None]  # (N, C, 1, 1)
        I, A, S = decompose_matrix(Y / norm)  # (N, C, 3, 3)

        # mix features again
        Y = self.W_I_post(I) + self.W_A_post(A) + self.W_S_post(S)
        return Y + torch.matrix_power(Y, 2)  # (N, C, 3, 3)


class ScalarOutput(nn.Module):
    """
    A non-linear read-out function:

    ``X`` with shape ``(N, C, 3, 3)`` is decomposed into ``I``, ``A``, and
    ``S`` components. The concatenation of the Frobenius norms of these
    components are passed through an MLP to generate a scalar.
    """

    def __init__(self, channels: int):
        super().__init__()
        self.layer_norm = nn.LayerNorm(3 * channels)
        self.mlp = MLP(
            layers=[3 * channels, 2 * channels, 1],
            activation=nn.SiLU(),
        )

    def forward(self, X: Tensor) -> Tensor:
        """X: (N, C, 3, 3) --> (N, 1)"""
        I, A, S = decompose_matrix(X)  # (N, C, 3, 3)
        norm_I = frobenius_norm(I)  # (N, C)
        norm_A = frobenius_norm(A)
        norm_S = frobenius_norm(S)

        X = torch.cat((norm_I, norm_A, norm_S), dim=-1)  # (N, 3C)
        X = self.layer_norm(X)
        return self.mlp(X)  # (N, 1)


class VectorOutput(nn.Module):
    """
    A non-linear read-out function:

    The ``A`` component of ``X`` with shape ``(N, C, 3, 3)`` is passed through
    a linear layer, before extracting the ``x``, ``y``, and ``z`` components
    of the resulting vector.
    """

    def __init__(self, channels: int):
        super().__init__()
        self.linear = TensorLinear(channels, 1)

    def forward(self, X: Tensor) -> Tensor:
        """X: (N, C, 3, 3) --> (N, 3)"""
        _, A, _ = decompose_matrix(X)  # (N, C, 3, 3)
        A_final = self.linear(A).squeeze()  # (N, 3, 3)
        # A_final[b, c] = [
        #     [  0, -vz,  vy],
        #     [ vz,   0, -vx],
        #     [-vy,  vx,   0],
        # ]
        x = A_final[..., 2, 1]
        y = A_final[..., 0, 2]
        z = A_final[..., 1, 0]
        return torch.stack((x, y, z), dim=-1)  # (N, 3)


### utils ###


def one_third_trace(X: Tensor) -> Tensor:
    """
    Calculate the one third trace of a (optionally batched) matrix, ``X``, of
    shape ``(...B, 3, 3)``.
    """
    return X.diagonal(offset=0, dim1=-1, dim2=-2).mean(-1)


def frobenius_norm(X: Tensor) -> Tensor:
    """
    Calculate the Frobenius norm of a (optionally batched) matrix, ``X``, of
    shape ``(...B, 3, 3)``.
    """
    return (X**2).sum((-2, -1))


def decompose_matrix(X: Tensor) -> tuple[Tensor, Tensor, Tensor]:
    """
    Take a (optionally batched) matrix, ``X``, of shape ``(...B, 3, 3)`` and
    decompose it into irreducible components, ``I``, ``A``, and ``S``:

    * ``I[b] = 1/3 * trace(X[b]) * Id``
    * ``A[b] = 0.5 * (X[b] - X[b].T)``
    * ``S[b] = 0.5 * (X[b] + X[b].T) - I[b]``

    where ``Id`` is the ``3x3`` identity matrix and ``b`` is some
    batch dimension(s)

    Parameters
    ----------
    x
        The matrix to decompose, of shape ``(...B, 3, 3)``.
    """
    trace = one_third_trace(X)[..., None, None]  # (...B, 1, 1)
    I = trace * torch.eye(3, 3, device=X.device, dtype=X.dtype)  # (...B, 3, 3)
    A = 0.5 * (X - X.transpose(-2, -1))
    S = 0.5 * (X + X.transpose(-2, -1)) - I
    return I, A, S  # 3x (...B, 3, 3)


def vector_to_skew_symmetric_matrix(v: Tensor) -> Tensor:
    """
    Creates a skew-symmetric tensor from a (optionally batched) vector.

    v: ([B], 3) --> sst: ([B], 3, 3)
    sst[b] = [
        [     0,   -v[b].z,   v[b].y],
        [ v[b].z,        0,  -v[b].x],
        [-v[b].y,   v[b].x,        0],
    ]

    """
    x, y, z = v.unbind(dim=-1)
    zero = torch.zeros_like(x)
    tensor = torch.stack(
        (
            zero,
            -z,
            y,
            z,
            zero,
            -x,
            -y,
            x,
            zero,
        ),
        dim=1,
    )  # (B, 9)
    return tensor.reshape(tensor.shape[:-1] + (3, 3))


def vector_to_symmetric_traceless_matrix(v: Tensor) -> Tensor:
    """
    Creates a symmetric traceless matrix from a vector.

    v: (..., 3) --> stm: (..., 3, 3)
    stm[...b] = 0.5 * (v[...b] v[...b].T) - 1/3 * trace(v[...b] v[...b].T) * Id
    """
    v_vT = torch.matmul(v.unsqueeze(-1), v.unsqueeze(-2))
    Id = torch.eye(3, 3, device=v_vT.device, dtype=v_vT.dtype)
    I = one_third_trace(v_vT)[..., None, None] * Id
    return 0.5 * (v_vT + v_vT.transpose(-2, -1)) - I


class TensorLinear(nn.Module):
    def __init__(self, in_features: int, out_features: int):
        super().__init__()
        self.linear = nn.Linear(in_features, out_features, bias=False)

    def forward(self, x: Tensor) -> Tensor:
        """x: (N, in, 3, 3) --> (N, out, 3, 3)"""
        return self.linear(x.transpose(-1, -3)).transpose(-1, -3)

    def __repr__(self):
        _in, _out = self.linear.in_features, self.linear.out_features
        return (
            f"{self.__class__.__name__}("
            f"[N, {_in}, 3, 3] --> [N, {_out}, 3, 3])"
        )



from __future__ import annotations

import os
from pathlib import Path

import pytorch_lightning as pl

from graph_pes.config.shared import instantiate_config_from_dict
from graph_pes.config.testing import TestingConfig
from graph_pes.models import load_model
from graph_pes.scripts.utils import (
    configure_general_options,
    extract_config_dict_from_command_line,
    update_summary,
)
from graph_pes.training.tasks import test_with_lightning
from graph_pes.utils import distributed
from graph_pes.utils.logger import logger


def test(config: TestingConfig) -> None:
    logger.info(f"Testing model at {config.model_path}...")

    configure_general_options(config.torch, seed=0)

    model = load_model(config.model_path)
    logger.info("Loaded model.")
    logger.debug(f"Model: {model}")

    datasets = (
        config.data if isinstance(config.data, dict) else {"test": config.data}
    )

    for dataset in datasets.values():
        if distributed.IS_RANK_0:
            dataset.prepare_data()
        dataset.setup()

    trainer = pl.Trainer(
        logger=config.get_logger(),
        accelerator=config.accelerator,
        inference_mode=False,
    )

    test_with_lightning(
        trainer,
        model,
        datasets,
        config.loader_kwargs,
        config.prefix,
        user_eval_metrics=[],
    )

    summary_file = Path(config.model_path).parent / "summary.yaml"
    update_summary(trainer.logger, summary_file)


def main():
    # set the load-atoms verbosity to 1 by default to avoid
    # spamming logs with `rich` output
    os.environ["LOAD_ATOMS_VERBOSE"] = os.getenv("LOAD_ATOMS_VERBOSE", "1")

    config_dict = extract_config_dict_from_command_line(
        "Test a GraphPES model using PyTorch Lightning."
    )
    _, config = instantiate_config_from_dict(config_dict, TestingConfig)
    test(config)


if __name__ == "__main__":
    main()
